/*
File selection screen for DS libfat

Copyright 2007 Damian Yerrick <pinoandchester@pineight.com>

This work is provided 'as-is', without any express or implied
warranty.  In no event will the authors be held liable for any
damages arising from the use of this work.

Permission is granted to anyone to use this work for any purpose,
including commercial applications, and to alter it and redistribute
it freely, subject to the following restrictions:

1. The origin of this work must not be misrepresented; you must
   not claim that you wrote the original work. If you use
   this work in a product, an acknowledgment in the product
   documentation would be appreciated but is not required.
2. Altered source versions must be plainly marked as such, and must
   not be misrepresented as being the original work.
3. This notice may not be removed or altered from any source
   distribution.

"Source" is the preferred form of a work for making changes to it.

*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <sys/syslimits.h>
#include <sys/unistd.h>
#include "fontdraw.h"
#include "xkeys.h"

#if 0
/**
 * Changes the name of the last thing after a / separator
 * to a dot.  Not used after the design of filesel was changed.
 */
void changeFileToDot(char *file) {
  char *slash = strrchr(path, '/');
  if (!dp && errno == ENOTDIR) {
    if (!slash || slash > path + (PATH_MAX - 3)) {
      vwfPuts(&vwfTouch, path, 0, 0);
      vwfPuts(&vwfTouch, "has no slash.", 0, 12);
      return -1;
    }
    slash[1] = '.';
    slash[2] = 0;
  }
}
#endif

typedef struct fileselEntry {
  char filename[NAME_MAX + 1];
  struct stat st;
} FileselEntry;

#define MAX_FILES_PER_FOLDER 256

static void printDir(FileselEntry *entries,
                     size_t nEntries,
                     int selected, int scroll) {
  int y = 12;
  for (int i = scroll;
       i < nEntries && i < scroll + 14;
       ++i) {
    const char *name = entries[i].filename;
    if (strcmp(name, "..") == 0) {
      name = "<Back>";
    }
    vwfRectfill(&vwfTouch,
                16, y,
                240, y + 12,
                selected == i ? 4 : 0);
    vwfPuts(&vwfTouch,
            S_ISDIR(entries[i].st.st_mode)
            ? "\002" : "\001",
            20, y);
    vwfPuts(&vwfTouch, name, 32, y);
    y += 12;
  }

  // Draw scroll bar  
  int yTop = scroll * 168 / nEntries;
  int yBottom = (nEntries > 14)
                ? (scroll + 14) * 168 / nEntries
                : 0;
  vwfRectfill(&vwfTouch,
              240, 12,
              256, yTop + 12,
              1);
  vwfRectfill(&vwfTouch,
              240, yTop + 12,
              256, yBottom + 12,
              2);
  vwfRectfill(&vwfTouch,
              240, yBottom + 12,
              256, 180,
              1);
}


#define FILESEL_CANCEL (-1)
#define FILESEL_OK 1
#define FILESEL_SCROLLING 2

int filesel(char *outName) {
  char oldcwd[PATH_MAX];
  int picked = 0;  // 1: chosen; 0: moving; -1: canceled
  FileselEntry *entries = malloc(sizeof(FileselEntry) * MAX_FILES_PER_FOLDER);
  int selected;  // index into entries
  
  u32 dasDir = 0;
  u32 dasCounter = 0;
  
  if (!entries) {
    return -1;
  }
  getcwd(oldcwd, sizeof(oldcwd));
  while (picked == 0) {
    DIR_ITER *dp = diropen(".");
    selected = 0;
    char cwd[PATH_MAX];
    getcwd(cwd, sizeof(cwd));

    if (dp) {
      size_t nEntries = 0;
      int lastJoy = ~0;
      int scroll = 0;
      int redraw = 1;
      
      // Read directory
      while (nEntries < MAX_FILES_PER_FOLDER
             && dirnext(dp,
                        entries[nEntries].filename,
                        &entries[nEntries].st) == 0) {
        if ((S_ISDIR(entries[nEntries].st.st_mode)
            && strcmp(entries[nEntries].filename, ".") != 0)
            || entries[nEntries].st.st_size == 262144) {
          ++nEntries;
        }
      }
	  dirclose(dp);
      
      vwfWinClear(&vwfTouch);
      vwfPuts(&vwfTouch, cwd, 8, 0);
      vwfPuts(&vwfTouch, "[ Cancel ]", 8, 180);
      vwfPuts(&vwfTouch, "[ OK ]", 136, 180);
      
      while (picked == 0 || picked == FILESEL_SCROLLING) {
      
        // auto scroll if the user isn't using the scroll bar
        if (picked != FILESEL_SCROLLING) {
          if (scroll > 0 && selected <= scroll) {
            scroll -= 1;
            redraw = 1;
          }
          if (scroll + 14 < nEntries && selected >= scroll + 13) {
            scroll += 1;
            redraw = 1;
          }
        }
        swiWaitForVBlank();
        if (redraw) {
          printDir(entries, nEntries, selected, scroll);
        }

        u32 j = readPad();
        u32 jnew = j & ~lastJoy;
        lastJoy = j;

        // handle Delayed Auto Shift (fixed at 250 ms 30 Hz)
        if (j & dasDir) {
          ++dasCounter;
          if (dasCounter >= 15) {
            dasCounter -= 2;
            jnew |= dasDir;
          }
        } else {
          dasCounter = 0;
        }

        if (jnew & KEY_UP) {
          dasDir = KEY_UP;
          if (selected > 0) {
            --selected;
            redraw = 1;
          }
        }
        if (jnew & KEY_DOWN) {
          dasDir = KEY_DOWN;
          if (selected < nEntries - 1) {
            ++selected;
            redraw = 1;
          }
        }
        if (jnew & KEY_A) {
          picked = FILESEL_OK;
        }
        if (jnew & KEY_B) {
          picked = FILESEL_CANCEL;
        }
        if (jnew & XKEY_TOUCH) {
          int tx = IPC->touchXpx, ty = IPC->touchYpx;
          int touchLine = ty / 12 - 1;

          if (touchLine == 14) {
          
            // Right side: OK; left side: Cancel.
            picked = (tx >= 128) ? 1 : -1;
          } else if (touchLine < 0) {

            // Touched directory name: Ignore.
          } else if (tx > 240) {

            // Touched scrollbar
            picked = FILESEL_SCROLLING;
          } else if (selected == touchLine + scroll) {

            // Touched the selected line: OK.
            picked = FILESEL_OK;
          } else if (touchLine + scroll < nEntries) {

            // Touched another line: Select it.
            selected = touchLine + scroll;
          }
        }
        if (picked == FILESEL_SCROLLING) {
          if (j & XKEY_TOUCH) {
            int ty = IPC->touchYpx;
            int relY = (ty - 12) * nEntries / 144;
            int scrollTarget = relY - 7;
            if (scrollTarget < 0) {
              scrollTarget = 0;
            } else if (scrollTarget > nEntries - 14) {
              scrollTarget = nEntries - 14;
            }
            if (scroll != scrollTarget) {
              scroll = scrollTarget;
              redraw = 1;
            }
          } else {
            picked = 0;
            if (selected < scroll) {
              selected = scroll;
            } else if (selected > scroll + 13) {
              selected = scroll + 13;
            }
          }
        }
      }

      // handle choosing a folder
      if (picked > 0 && S_ISDIR(entries[selected].st.st_mode)) {
        chdir(entries[selected].filename);
        picked = 0;
      }
    } else {
      vwfWinClear(&vwfTouch);
      vwfPuts(&vwfTouch, "diropen failed in", 0, 288);
      vwfPuts(&vwfTouch, cwd, 12, 288);
      vwfPuts(&vwfTouch, strerror(errno), 24, 288);
      picked = -1;
    }
  }

  if (picked > 0) {

    // return the file name
    strncpy(outName, entries[selected].filename, NAME_MAX);
    outName[NAME_MAX] = 0;
    free(entries);
    return 0;
  } else {
    chdir(oldcwd);
    free(entries);
    return picked;
  }
}
