#include <stdio.h>
#include "acstr.h"
#include "fontdraw.h"
#include <string.h>
char *stpcpy(char *restrict dst, const char *restrict src);

#define CALC_NOFFSET_E(in) (0x9094 + 0x700 * (in))
const char *getNeighborName(size_t id);

/**
 * Calculates the sum of all little-endian 2-byte words in a buffer.
 * @param data pointer to the start of the buffer
 * @param len length of the buffer in bytes
 * @return the sum
 */
u32 u16sum(const unsigned char *data, size_t len) {
  u32 sum = 0;
  
  for (len /= sizeof(u16); len > 0; --len) {
    sum += *data++;
    sum += *data++ << 8;
  }
  return sum;
}

/**
 * Guesses whether a save file is an Animal Crossing town.
 * Heuristic: In a valid town, the town name should match between
 * the primary copy and the backup copy in the save file.
 * @param acwwData pointer to a 262144 byte save file
 * @return 0 for not an ACWW town, 1 for probably an ACWW town
 */
int isACWWTown(const unsigned char *acwwData) {
  return memcmp(acwwData+2, acwwData+2+0x15FE0, 10) == 0;
}

/**
 * Displays information about a loaded Animal Crossing town.
 */
void acTownInfo(const unsigned char *acwwData) {
  int y = 0;
  char line[64];
  char *cursor;

  vwfWinClear(&vwfTouch);

  // Show checksum
  siprintf(line,
           "Checksum: %08x (should end in 0000)",
           u16sum(acwwData, 0x15FE0));
  vwfPuts(&vwfTouch, line, 0, y);

  // Show town name  
  cursor = stpcpy(line, "ACWW town: ");
  acstrDecode(cursor, acwwData+4, 8);
  vwfPuts(&vwfTouch, line, 0, y += 12);
  
  vwfPuts(&vwfTouch, "Residents:", 0, y += 12);
  y += 12;

  // Display player names  
  int x = 0;
  for (int i = 0; i < 4; i++) {
    const unsigned char *name = acwwData + 0x228E + 0x228C * i;
    if (*name) {
      acstrDecode(line, name, 8);
      vwfPuts(&vwfTouch, line, x, y);
      x += 64;
    }
  }

  // Display neighbor names  
  x = 256;
  for (int i = 0; i < 8; ++i) {
    int identity = acwwData[CALC_NOFFSET_E(i) + 0x73];
    const char *name = getNeighborName(identity);
    if (identity < 150) {
      if (x >= 256) {
        x = 0;
        y += 12;
      }
      vwfPuts(&vwfTouch, name, x, y);
      x += 64;
    }
  }

}

typedef struct ACWWName {
  u16 id;
  ACSTR name[8];
} ACWWName;




#if 0

/* World map */

typedef u16 ACWW_MAP_ACRE[16][16];
typedef ACWW_MAP_ACRE ACWW_MAP[4][4];

ACWW_MAP *mapData = (ACWW_MAP *)(acwwSave + 0x00C354);

/* Patterns */


typedef struct ACWWPattern {
  u8 patData[32][16];
  ACWWName authorTown;
  ACWWName author;
  u16 reserved1;
  ACSTR patName[16];
  u16 reserved2;
} ACWWPattern;

#define INV_PATTERNS(player) \
  ((ACWWPattern *)(acwwSave + 0x00000C + player * 0x228C))

/* Letters to be sent */
typedef ACSTR LetterInPO[0xC4];
#define LETTERS_IN_PO \
  ((LetterInPO *)(acwwData + 0x011DB4))

#endif

/* Letters *********************************************************/

#define INV_LETTERS(player) \
  ((LetterInInv *)(acwwData + 0x001154 + player * 0x228C))
#define SAVED_LETTERS(player) \
  ((LetterInInv *)(acwwData + 0x02E20C + player * 0xF4 * 75))

/*
Letters from neighbors will have "" as the value of nameFrom.name.
Instead, it is somehow encoded in nameFrom.id.
A letter from Chevre (neighbor species #0x48) in Picken is 0x4803@YoshiX.
A letter to Chow was 0x0702.
Speculation: The townFrom for a neighbor is based on the where the
neighbor was "born".

Letters from Mom are ""(0000)@""(0000). So are autogen bottlemails.
Letters from Nintendo are Nintendo(81D4)@Redmond(C1BD)

Empty slots are entirely 0'd, except for attached item = empty slot.

The "To" and "From" addresses determine
It appears that the separate are based on
Unk1[2], Unk2[2], flags, specialSender
which I've called "address type data".

Unk1[2] determines the type of the "To" address
Unk2[2] determines the type of the "From" address
  0 empty inv slot
  1 your future self (To only?)
  2 a player (name is in nameFrom.name)
  3 a neighbor (name is zeroed; type in nameFrom.id >> 8)
  4 a special character (From only?)
  6 autogen bottlemail (To only?)
  7 human bottlemail (To only?)
flags:
  0x00 empty inv slot
  0x01 Composing
  0x02 Unread letter
  0x03 Read letter
  0x05 Unread bottlemail
  0x06 Read bottlemail
  +0x40 If item is attached, it is unopened
specialSender:
  0 player, neighbor, or empty inv slot
  1 Mom
  2 HRA
  7 Lyle
  11 Redd
  12 Autogen bottlemail
  15 Tortimer
  17 Wi-Fi letter from Nintendo

Position of salutation appears related to Unk3[0]
Chester, your room gets : salNamePos = 0
Dear Chester, : salNamePos = 5
Smarty-pants Chester : salNamePos = 13
As far as I can tell, Players, neighbors, and some specialSenders
use a salutation; other specialSenders don't.

*/
typedef struct LetterInInv {
  u8 unknown0[4];
  ACWWName townTo;
  ACWWName nameTo;
  u8 unknown1[4];
  ACWWName townFrom;
  ACWWName nameFrom;
  u8 unknown2[4];
  ACSTR salutation[24];
  ACSTR body[128];
  ACSTR closing[32];
  u8 salNamePos;
  u8 paperType;
  u8 flags;
  u8 specialSender;
  u16 item;
  u8 unknown4[2];
} LetterInInv;

static const char hexDigits[] = "0123456789abcdef";

void drawHexByte(int x, unsigned int row, int n) {
  int y = row * 12;
  unsigned int hiNibble = hexDigits[(n & 0xF0) >> 4];
  unsigned int loNibble = hexDigits[(n & 0x0F)];
  vwfPutc(&vwfTouch, hiNibble, x, y);
  vwfPutc(&vwfTouch, loNibble, x + 6, y);
}

#define HEX_BYTES_PER_ROW 16
#define HEX_ROWS_PER_SCREEN 16
#define HEX_PIXELS_PER_BYTE 14
#define HEX_LEFT 32
void dispHex(const unsigned char *data, size_t n) {
  vwfWinClear(&vwfTouch);
  if (n > HEX_BYTES_PER_ROW * HEX_ROWS_PER_SCREEN) {
    n = HEX_BYTES_PER_ROW * HEX_ROWS_PER_SCREEN;
  }
  for (size_t row = 0; n > 0; ++row) {
    drawHexByte(8, row, row * HEX_BYTES_PER_ROW);
    for (size_t x = 0;
         n > 0 && x < HEX_BYTES_PER_ROW;
         --n, ++x) {
      drawHexByte(x * HEX_PIXELS_PER_BYTE + HEX_LEFT, row, *data++);
    }
    data += HEX_BYTES_PER_ROW;
  }
}

static const char *const specialSenders[] = {
  [1] = "Mom",
  [2] = "Happy Room Academy",
  [3] = "Tom Nook",
  [6] = "Pelly",
  [7] = "Lyle",
  [9] = "Wishy",
  [11] = "Redd",
  [12] = "Bottle",
  [15] = "Tortimer",
  [17] = "Nintendo"
};

static char *letterAddrDecode(char *cursor, const ACWWName *town,
                              const ACWWName *name, unsigned int type,
                              unsigned int specialSender) {
  switch (type) {
  case 0:
    break;

  case 1:  // future self
  case 2:  // player
    cursor = acstrDecode(cursor, name->name, 8);
    cursor += siprintf(cursor, "(%04x)@", name->id);
    cursor = acstrDecode(cursor, town->name, 8);
    cursor += siprintf(cursor, "(%04x)", town->id);
    break;
  
  case 3:  // neighbor
    {
      const char *neighborName = getNeighborName(name->id >> 8);
      cursor = stpcpy(cursor, neighborName ? neighborName : "Unknown");
      cursor += siprintf(cursor, "(%04x)@", name->id);
      cursor = acstrDecode(cursor, town->name, 8);
      cursor += siprintf(cursor, "(%04x)", town->id);
    }
    break;
    
  case 4:  // special character
    {
      const char *name = specialSender < 18
                         ? specialSenders[specialSender]
                         : NULL;
      if (name) {
        cursor = stpcpy(cursor, name);
      } else {
        cursor += siprintf(cursor, "Visitor %u", specialSender);
      }
    }
    break;

  case 6:  // autogen bottlemail
  case 7:  // player bottlemail
    cursor = stpcpy(cursor, "Some Stranger");
  }
  
  *cursor = 0;
  return cursor;
}


void dispLetter(const LetterInInv *letter) {
  char line[100];

  if (sizeof(*letter) != 0xF4) {
    vwfPuts(&vwfTouch, "Size of letter is incorrect!", 4, 0);
    return;
  }

  // Display address (will show up as 0's if invalid)
  char *cursor = stpcpy(line, "To ");
  cursor = letterAddrDecode(cursor,
                            &letter->townTo, &letter->nameTo,
                            letter->unknown1[2], letter->specialSender);
  vwfPuts(&vwfTouch, line, 4, 24);

  cursor = stpcpy(line, "From ");
  cursor = letterAddrDecode(cursor,
                            &letter->townFrom, &letter->nameFrom,
                            letter->unknown2[2], letter->specialSender);
  vwfPuts(&vwfTouch, line, 4, 36);

  // Display salutation, body, closing
  acstrDecode(line, letter->salutation, sizeof(letter->salutation));
  vwfPuts(&vwfTouch, line, 4, 48);
  for (int y = 0; y < sizeof(letter->body) / 32; ++y) {
    int start = y * 32;
    int end = start + 32;
    int len = end - start;

    acstrDecode(line, letter->body + start, len);
    vwfPuts(&vwfTouch, line, 4, 60 + y * 12);
  }
  acstrDecode(line, letter->closing, sizeof(letter->closing));
  vwfPuts(&vwfTouch, line, 4, 108);

  // If non-empty item, display it.
  if (letter->item != 0xFFF1) {
    siprintf(line, "Item: %04x", letter->item);
    vwfPuts(&vwfTouch, line, 4, 180);
  } else {
    vwfPuts(&vwfTouch, "No item", 4, 180);
  }

  siprintf(line, "Address type data: %d, %d, %02x, %d",
           letter->unknown1[2],
           letter->unknown2[2],
           letter->flags,
           letter->specialSender);
  vwfPuts(&vwfTouch, line, 4, 144);

  siprintf(line, "Sal: %d; Paper: %02x",
           letter->salNamePos,
           letter->paperType);
  vwfPuts(&vwfTouch, line, 4, 156);

#if 0
  // Display fields whose significance is unknown as hex
  siprintf(line, "Unk0: %02x %02x %02x %02x",
           letter->unknown0[0],
           letter->unknown0[1],
           letter->unknown0[2],
           letter->unknown0[3]);
  vwfPuts(&vwfTouch, line, 128, 156);
  siprintf(line, "UnkTo: %02x %02x %02x %02x",
           letter->unknown1[0],
           letter->unknown1[1],
           letter->unknown1[2],
           letter->unknown1[3]);
  vwfPuts(&vwfTouch, line, 4, 168);
  siprintf(line, "UnkFrom: %02x %02x %02x %02x",
           letter->unknown2[0],
           letter->unknown2[1],
           letter->unknown2[2],
           letter->unknown2[3]);
  vwfPuts(&vwfTouch, line, 128, 168);
  siprintf(line, "Unk4: %02x %02x",
           letter->unknown4[0],
           letter->unknown4[1]);
  vwfPuts(&vwfTouch, line, 128, 180);
#endif
}

void dispLetterInInv(const ACSTR *acwwData, size_t player, size_t slot) {
  const LetterInInv *letter = INV_LETTERS(player)+slot;

  vwfWinClear(&vwfTouch);
  {
    char line[64];
    siprintf(line, "Letter in player %d inventory slot %d", player + 1, slot + 1);
    vwfPuts(&vwfTouch, line, 4, 0);
  }
  dispLetter(letter);
}

void dispSavedLetter(const ACSTR *acwwData, size_t player, size_t slot) {
  const LetterInInv *letter = SAVED_LETTERS(player)+slot;

  vwfWinClear(&vwfTouch);
  {
    char line[64];
    siprintf(line, "Letter in player %d deposit box %d", player + 1, slot + 1);
    vwfPuts(&vwfTouch, line, 4, 0);
  }
  dispLetter(letter);
}


/**
 * Guesses whether a save file is for Drawn to Life (U).
 * Heuristic: First four bytes are LOTD ("Drawn TO Life" backwards).
 * @param acwwData pointer to a 262144 byte save file
 * @return 0 for not DTOL, >0 for probably DTOL
 */
int isDTOL(const unsigned char *dtolData) {
  return !memcmp("LOTD", dtolData, 4);
}

void getDTOLName(char *dst, const unsigned char *dtolData) {
  memcpy(dst, "DTOL_", 5);
  memcpy(dst + 5, dtolData + 0x7914, 12);
  dst[5 + 12] = 0;
}

/**
 * Guesses whether a save file is for March of the Minis (U).
 * Heuristic: First four bytes are LOTD ("Drawn TO Life" backwards).
 * @param acwwData pointer to a 262144 byte save file
 * @return 0 for not DTOL, >0 for probably DTOL
 */
int isMOTM(const unsigned char *motmData) {
  return !memcmp("MUNCHKIN", motmData, 8);
}

void getMOTMName(char *dst, const unsigned char *motmData) {
  memcpy(dst, "MOTM_", 5);
  for (int x = 0; x < 10; x++) {
    int lobyte = motmData[0x6A12 + 2 * x];
    int hibyte = motmData[0x6A13 + 2 * x];
    dst[x + 5] = (hibyte ? '_' : motmData[0x6A12 + 2 * x]);
  }
  dst[5 + 10] = 0;
}

/*

Future games

Pokemon Mystery Dungeon Blue Rescue Team
detect: POKE_DUNGEON at 0x0404 and 0x6404
but no known town- or player-identifying marks

Advance Wars Dual Strike
detect: dswj at each 0x3000 bytes from 0x00000 to 0x3C000
but no known town- or player-identifying marks

*/
